#include "ScoreA1.h"

ScoreA1::ScoreA1(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreA1::ScoreA1]  ";

  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  //ioh.openInput("genofile", &genoStream);
  //dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new int[nsamples];
  globalAnc = new double[nsamples];

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  if ( ioh.getBoolParam("oddsratio") )  {
    ioh.openOutput("orfile",&oddsratioStream);
  }

  // ioh.dbgShowParams();

  dipLocalAnc = new int[nsamples];
  do_not_deallocate = 0;

  return;
}

ScoreA1::ScoreA1(int n, int *p, int *a, double *t)  {
  do_not_deallocate = 1;
  nsamples = n; 
  dipPhenoVec = p;
  dipLocalAnc = a;
  globalAnc = t;
}

ScoreA1::~ScoreA1(void)  {

  if ( do_not_deallocate )  
    return;

  // close file streams 
  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( oddsratioStream.is_open() )  {  oddsratioStream.close();  } 
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )  {  localAncStream.close();  }
  // mem: global anc stream already closed

  // free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete globalAnc;
  delete dipLocalAnc;
}

int ScoreA1::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreA1::evaluate]  ";
  
  bool output_to_file =ioh.getBoolParam("output2file");
  bool output_oddsratios = (ioh.getBoolParam("oddsratio") &&
         !do_not_deallocate);  // i.e., if not running through MIX

  try { ioh.readIntLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << "error reading phenotype" << endl;
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  ioh.getParam("thetafile", &str);
  try { ioh.readOneColumnFile(str, nsamples, &globalAnc); }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage();
    throw( Exception (errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalA1();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_oddsratios )  {
      oddsratioStream << fixed << setw(10) << setprecision(4) << 
        oddsratio1 << endl;
    }
  }

  return 0;
}

#define RMIN 0.1
#define RMAX 10.0

#define NO_CONVERGE 2
#define ZERO_DERIV 1

#define VERYSMALL 1e-8
#define MAXITER 40



double ScoreA1::evalA1(void)  {
  ostringstream errstream;
  errstream << "[ScoreA1::evalA1]  ";
  
  initializeDip();

  int ierr = 0;
  double nullscore = f(1.0);
  double R0 = find_initial_rval();
  double radmix;
  
  try  {
    radmix = findroot(R0,&ierr);
  }
  catch ( Exception& e )  {
    errstream << "error finding root" << endl;
    errstream << e.getMessage();
    throw( Exception(errstream.str()) );
  }
  setOddsRatios(radmix);
  double causalscore = f(radmix);
  finalizeDip();
  return( causalscore - nullscore );

}

// ------------------------------------------------------------------------
void ScoreA1::initializeDip()  {
  AS = new int[nsamples];  
  TS = new double[nsamples];
  // fill case-only arrays
  int j = 0;
  ncase = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( dipPhenoVec[i] == 1 )  {
      AS[j] = dipLocalAnc[i];
      TS[j] = globalAnc[i];
      ncase += 1;
      j++;
    }
  }
  
  // count local ancestry incidence in case-only arrays
  // (needed for computation of derivative;
  nee = 0.0;
  nae = 0.0;
  for(int j=0;j<ncase;j++)  {
    if ( AS[j] == 1 )
      nae += 1;
    else if ( AS[j] == 2 )
      nee += 1;
  }
}

// ------------------------------------------------------------------------
void ScoreA1::finalizeDip()  {
  delete AS;
  delete TS;
}

// ------------------------------------------------------------------------
double ScoreA1::findroot(double rinitial, int *ierr)  {
  ostringstream errstream;
  errstream << "[ScoreA1::findroot]  ";

  int myitcount = 0;
  double delta;
  *ierr = 0;
  double myr = rinitial;

  do  {

    // find root of DERIVATIVE
    double farg = df(myr);
    double dfarg = ddf(myr);

    if ( fabs(dfarg) < VERYSMALL )  {
      errstream << "derivative zero"; 
      throw(Exception(errstream.str()));
    }

    if ( myitcount++ > MAXITER )  {
      errstream << "no convergence in" << (myitcount-1) << "iterations"; 
      throw(Exception(errstream.str()));
    }
  
    myr -= (farg/dfarg);
    delta = fabs(farg/dfarg);

  } while ( fabs(delta) > VERYSMALL );
  return ( myr );

}

// ------------------------------------------------------------------------
double ScoreA1::f(double r)  {

  double sum = 0.0;
  for(int i=0;i<ncase;i++)  {
    double a1 = TS[i]*TS[i]*r*r;
    double a2 = 2*TS[i]*(1.0-TS[i])*r;
    double a3 = (1.0-TS[i])*(1.0-TS[i]);
    double b = a1+a2+a3;

    if ( AS[i] == 2 )
      sum += (log(a1) - log(b));
    else if ( AS[i] == 1 )
      sum += (log(a2) - log(b));
    else if ( AS[i] == 0 )
      sum += (log(a3) - log(b));
  }

  return (2*sum);
}

// ------------------------------------------------------------------------
double ScoreA1::df(double r)  {
  double sum = (2.0*nee + nae)/r;
  for(int i=0;i<ncase;i++)  {
    double x = 2.0*r*TS[i]*TS[i] + 2.0*TS[i]*(1.0-TS[i]);
    double y = TS[i]*TS[i]*r*r + 2.0*TS[i]*(1.0-TS[i])*r + (1.0-TS[i])*(1.0-TS[i]);
    sum -= (x/y);
  }
  return sum; 
}

// ------------------------------------------------------------------------
double ScoreA1::ddf(double r)  {
  double sum = -(2.0*nee + nae)/(r*r);
  for(int i=0;i<ncase;i++)  {
    double x = 2.0*r*TS[i]*TS[i] + 2.0*TS[i]*(1.0-TS[i]);
    double y = TS[i]*TS[i]*r*r + 2.0*TS[i]*(1.0-TS[i])*r + (1.0-TS[i])*(1.0-TS[i]);
    double dx = 2.0*TS[i]*TS[i];
    sum -= (dx/y - (x*x)/(y*y));
  }
  return sum;
}

// ------------------------------------------------------------------------
double ScoreA1::find_initial_rval(void)  {
  double maxf = -1.0e10, maxr = 1.0;
  for(double r=0.10;r<=10.0;r+=0.05)  {
    double fval = f(r);
    if ( fval > maxf )  {
      maxf = fval;
      maxr = r;
    }
  }
  return maxr;

}

// ------------------------------------------------------------------------
void ScoreA1::setOddsRatios(double r)  {
  
  if ( r <= RMIN )  {
    oddsratio1 = "small";
  }
  else if ( r >= RMAX )  { 
    oddsratio1 = "large";
  }
  else  {
    ostringstream rstr;
    rstr << fixed << setw(10) << setprecision(4) << r;
    oddsratio1 = rstr.str();
  }

  // For now, don't compute R_snp
  oddsratio2 = "not-used";

}

