#include "ScoreA2Q.h"

ScoreA2Q::ScoreA2Q(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreA2Q::ScoreA2Q]  ";

  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new double[nsamples];
  globalAnc = new double[nsamples];

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  // ioh.dbgShowParams();

  // allocate memory 
  dipPhenoVec  = new double[nsamples];
  globalAnc    = new double[nsamples];

  dipLocalAnc = new int[nsamples];
  do_not_deallocate = 0;

  return;
}

ScoreA2Q::ScoreA2Q(int n, double *p, int *a, double *t)  {
  do_not_deallocate = 1;
  nsamples = n;
  dipPhenoVec = p;
  dipLocalAnc = a;
  globalAnc = t;
}

ScoreA2Q::~ScoreA2Q(void)  {
  string str;

  if ( do_not_deallocate )
    return;

  // close file streams 
  if ( outputStream.is_open() ) { outputStream.close(); }
  if ( phenoStream.is_open()  ) { phenoStream.close();  }
  if ( localAncStream.is_open() )  {  localAncStream.close();  }
  // mem: global anc stream already closed

  // free memory
  delete dipPhenoVec;
  delete globalAnc;
  delete dipLocalAnc;

}

int ScoreA2Q::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreA2Q::evaluate]  ";
  
  bool output_to_file =ioh.getBoolParam("output2file");

  try { ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << "error reading phenotype" << endl;
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  ioh.getParam("thetafile", &str);
  try { ioh.readOneColumnFile(str, nsamples, &globalAnc); }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage();
    throw( Exception (errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalA2Q();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }
  }

  return 0;
}

double ScoreA2Q::evalA2Q(void)  {

  ostringstream errstream;
  errstream << "[ScoreA2Q::evalA2Q] ";

  int *baddata = new int[nsamples];
  double *alpha = new double[nsamples];

  int count;
  double sum;

  count = 0;
  for(int i=0;i<nsamples;i++)  {
    baddata[i] = 0;
    if ( fabs(dipPhenoVec[i]+1024) < 1e-6 )   // "missing"
      baddata[i] = 1;
    count++;
  }

  // barely possible
  if ( count == 0 )  {
    errstream << "no valid data";
    throw(Exception(errstream.str()));
  }


  // real-valued local ancestry
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )  
      continue;
    alpha[i] = (double) dipLocalAnc[i];
  }

  // mean-adjust global ancestry
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    sum += globalAnc[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    globalAnc[i] -= (sum/count);
  }

  // adjust phenotype for global ancestry
  double num = 0.0;
  double den = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    num += dipPhenoVec[i]*globalAnc[i];
    den += globalAnc[i]*globalAnc[i];
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    dipPhenoVec[i] -= (num/den)*globalAnc[i];
  }

  // adjust local ancestry for global ancestry
  num = 0.0;
  den = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    num += alpha[i]*globalAnc[i];
    den += globalAnc[i]*globalAnc[i];
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    alpha[i] -= (num/den)*globalAnc[i];
  }
  
  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  int sum1 = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    sumx += dipPhenoVec[i];
    sumy += alpha[i];
    sumxx += (dipPhenoVec[i]*dipPhenoVec[i]);
    sumxy += (dipPhenoVec[i]*alpha[i]);
    sumyy += (alpha[i]*alpha[i]);
    sum1  += 1;
  }

  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1 - mx*mx);
  double sy = sqrt(sumyy/sum1 - my*my);
  if ( sx < 1e-8 || sy < 1e-8 )  {
    errstream << "zero variance";
    throw(Exception(errstream.str()));
  }

  double corr = (sumxy/sum1 - mx*my)/(sx*sy);

  delete baddata;
  delete alpha;

  return ( sum1*corr*corr );

}





















