#include "ScoreHQ.h"

ScoreHQ::ScoreHQ(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreHQ::ScoreHQ]  ";

  // read parameter file
  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new double[nsamples];

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  if ( ioh.getBoolParam("effectsize") )  {
    ioh.openOutput("esfile",&effectsizeStream);
  }

  // ioh.dbgShowParams();

  return;
}

ScoreHQ::~ScoreHQ(void)  {

  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( effectsizeStream.is_open() ) {  effectsizeStream.close(); }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )   {  localAncStream.close();  }

  // close file streams and free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete dipLocalAnc;
}

int ScoreHQ::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreHQ::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 
  bool output_effectsizes = ioh.getBoolParam("effectsize"); 

  try { ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }   
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading genotype" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double sh, s2q;
    try  {
      sh = evalHQ();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    try  {  
      ScoreS2Q S2Q(nsamples, dipGenoVec, dipPhenoVec, dipLocalAnc);
      s2q = S2Q.evalS2Q();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();
      cerr << errstream.str();
      if ( output_to_file )  {  
        outputStream << "no-score" << endl;
      }
      continue;
    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << (sh-s2q) << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << (sh-s2q) << endl;
    }

    if ( output_effectsizes )  {
      effectsizeStream << fixed << setw(10) << setprecision(4) << 
        effectsize1 << "  " << effectsize2 << endl;
    }
  }

  return 0;
}



#define TINY 0.001
#define MINP 0.0001
#define MAXP 0.9990
#define VERY_SMALL 1e-8
#define MAX_ITER 40

#define DBG 0

/* ------------------------------------------------------------------- */
/*                              Score H                                */
/* ------------------------------------------------------------------- */
double ScoreHQ::evalHQ(void)  {

  ostringstream errstream;
  errstream << "[ScoreHQ::evalHQ]  ";

  int count;
  double sum;

  bool hsimple_eval = ioh.getBoolParam("hsimple");

  countAll(dipGenoVec, dipLocalAnc);

  /* find pA and pE by maximizing likelihood */
  double pA = initpA();
  double pE = initpE();

  double A0 = 2* XAARR + XAARV + XAERR;
  double B0 = 2* XAAVV + XAARV + XAEVV;
  double C0 = 2* XEERR + XAERR + XEERV;
  double D0 = 2* XEEVV + XEERV + XAEVV;
  double X0 = XAERV;

  int done =  0;
  int itercount = 0;
  while ( !done && itercount++ < MAX_ITER )  {
    double oldpA = pA, oldpE = pE;
    pA = rootpA(A0,B0,C0,D0,X0,pE);
    pE = rootpE(A0,B0,C0,D0,X0,pA);
    done = ( fabs(pA-oldpA) < VERY_SMALL && 
             fabs(pE-oldpE) < VERY_SMALL );
  }
  if ( itercount > MAX_ITER )  {
    errstream << "failure to converge (can't find pA and pE)";
    throw(Exception(errstream.str()));
  }
  if ( DBG ) { 
     cout << "pA " << fixed << setw(10) << setprecision(4) << pA  << "  " 
          << "pE " << pE << endl;
  }

  double *pheno = new double[nsamples];  
  double *genoA = new double[nsamples];  
  double *genoE = new double[nsamples];  
  double *gamma = new double[nsamples];  
  int *baddata = new int[nsamples];

  // find bad data
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    baddata[i] = 0;
    if ( fabs(dipPhenoVec[i]+1024.0) < 1e-6 )
      baddata[i] = 1;
    if ( dipGenoVec[i] == 9 )  
      baddata[i] = 1;
    if ( baddata[i] == 0 )
      count++;
  }

  /* mean-adjust phenotype */
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      sum += dipPhenoVec[i];
      count++;
    }
  }
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      pheno[i] = dipPhenoVec[i] - sum/count;
    }
  }


  /* make genoA and genoE */
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    if ( dipLocalAnc[i] == 0 )  {
      genoA[i] = dipGenoVec[i];
      genoE[i] = 0.0;
    }
    else if ( dipLocalAnc[i] == 1 )  {
      if ( hsimple_eval )  {
        genoA[i] = 0.5*dipGenoVec[i];
        genoE[i] = 0.5*dipGenoVec[i];
      }
      else  {
        genoA[i] = dipGenoVec[i]*pA/(pA+pE);
        genoE[i] = dipGenoVec[i]*pE/(pA+pE);
      }
    }
    else if ( dipLocalAnc[i] == 2 )  {
      genoA[i] = 0.0;
      genoE[i] = dipGenoVec[i];
    }
  }

  /* mean-adjust genoA */
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    sum += genoA[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    genoA[i] -= (sum/count);
  }

  /* mean-adjust genoE */
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    sum += genoE[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    genoE[i] -= (sum/count);
  }

  /* mean-adjust local ancestry */
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    double a = dipLocalAnc[i];  
    sum +=a;
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    double a = dipLocalAnc[i];  
    gamma[i] = a - (sum/count);
  }


  /* adjust genoA for gamma */
  double numer = 0.0;
  double denom = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    numer += genoA[i]*gamma[i];
    denom += gamma[i]*gamma[i];
  }
  if ( DBG ) { cout << "gamma predicts genoA: " << fixed << setw(10) << setprecision(4) << (numer/denom) << endl; }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    genoA[i] -= (numer/denom)*gamma[i];
  }


  /* adjust genoE for gamma */
  numer = 0.0;
  denom = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    numer += genoE[i]*gamma[i];
    denom += gamma[i]*gamma[i];
  }
  if ( DBG ) { cout << "gamma predicts genoE: " << fixed << setw(10) << setprecision(4) << (numer/denom) << endl; }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    genoE[i] -= (numer/denom)*gamma[i];
  }


  /* adjust pheno for gamma */
  numer = 0.0;
  denom = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    numer += pheno[i]*gamma[i];
    denom += gamma[i]*gamma[i];
  }
  if ( DBG ) { cout << "gamma predicts pheno: " << fixed << setw(10) << setprecision(4) << (numer/denom) << endl; }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    pheno[i] -= (numer/denom)*gamma[i];
  }

  /* two-dimensional regression */
  double X11 = 0.0;
  double X12 = 0.0;
  double X22 = 0.0;
  double Y1 = 0.0;
  double Y2 = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    X11 += (genoA[i]*genoA[i]);
    X12 += (genoA[i]*genoE[i]);
    X22 += (genoE[i]*genoE[i]);
    Y1 += genoA[i]*pheno[i];
    Y2 += genoE[i]*pheno[i];
  }
  double det = (X11*X22-X12*X12);

  double V11 = X22/det;
  double V22 = X11/det;
  double V12 = -X12/det;

  double epsA = V11*Y1 + V12*Y2;
  double epsE = V12*Y1 + V22*Y2;
  if ( DBG ) { 
     cout << "epsA " << fixed << setw(10) << setprecision(4) << epsA  << "  "
          << "epsE " << epsE << endl;
  }

  setEffectSizes(epsA, epsE);

  /* compute total variance and predicted variance */
  count = 0;
  double totvar = 0.0;
  double predvar = 0.0;
  for(int i=0;i<nsamples;i++)  {
    double pred;
    if ( baddata[i] ) 
      continue;
    totvar += (pheno[i]*pheno[i]);
    pred = epsA*genoA[i] + epsE*genoE[i];
    predvar += (pred*pred);
    count++;
  }

  delete pheno;
  delete genoA;
  delete genoE;
  delete gamma;
  delete baddata;

  if ( DBG ) { cout << "result = " << fixed << setw(10) << setprecision(4) << count*predvar/totvar << endl; }

  return( count*predvar/totvar );

}


void ScoreHQ::countAll(int *G, int *A)  {

  XAAVV = XAARV = XAARR = TINY;
  XAEVV = XAERV = XAERR = TINY;
  XEEVV = XEERV = XEERR = TINY;

  for(int i=0;i<nsamples;i++)  {
    if ( G[i] == 9 || A[i] == 9 )  
      continue;

    if ( A[i] == 0 )  {
      if ( G[i] == 0 )  
        XAAVV += 1;
      if ( G[i] == 1 ) 
        XAARV += 1;
      if ( G[i] == 2 ) 
        XAARR += 1;
    }
    if ( A[i] == 1 )  {
      if ( G[i] == 0 )  
        XAEVV += 1;
      if ( G[i] == 1 ) 
        XAERV += 1;
      if ( G[i] == 2 ) 
        XAERR += 1;;
    }
    if ( A[i] == 2 )  {
      if ( G[i] == 0 )  
        XEEVV += 1;
      if ( G[i] == 1 ) 
        XEERV += 1;
      if ( G[i] == 2 ) 
        XEERR += 1;
    }
  } 
}

double ScoreHQ::rootpA(double A, double B, double C, double D, double X, double pE)  {

  double a = -(A+B+X)*(1-2*pE);
  double b = A*(1-3*pE) - B*pE + X*(1-2*pE);
  double c = A*pE;
  double pA;

  if ( fabs(a) < VERY_SMALL && fabs(b) < VERY_SMALL )  {
    errflag = 1;
    return(-1);
  }

  if ( fabs(a) < VERY_SMALL)  {
    pA = -(c/b);
  }
  else  { 
    double disc = b*b-4*a*c;
    if ( disc < 0 )  {
      /* check endpoints for maximum */
      double s0, s1;
      s0 = S(A,B,C,D,X,MINP,pE);
      s1 = S(A,B,C,D,X,MAXP,pE);
      pA = ( s0 < s1 ? MAXP : MINP );
    }
    else if ( disc > 0 )  {

      double r1 = (-b+sqrt(disc))/(2*a);
      double r2 = (-b-sqrt(disc))/(2*a);

      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP  > r2 || r2  > MAXP ) )  {
        pA = r1;
      }
      if ( ( MINP <= r2 && r2 <= MAXP ) && 
           ( MINP  > r1 || r1  > MAXP ) )  {
        pA = r2;
      }
      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP <= r2 && r2 <= MAXP ) )  {
        double d1 = d2S(A,B,X,r1,pE);
        double d2 = d2S(A,B,X,r2,pE);
        if ( d1 < 0 && d2 >= 0 )  {
          pA = r1;
        }
        else if ( d2 < 0 && d1 >= 0 )  {
          pA = r2;
        }
        else  {
          double s0 = S(A,B,C,D,X,MINP,pE);
          double s1 = S(A,B,C,D,X,MAXP,pE);
          pA = ( s0 < s1 ? MAXP : MINP );
        }
      }
      if ( ( MINP > r1 || r1 > MAXP ) && 
           ( MINP > r2 || r2 > MAXP ) )  {
        double s0 = S(A,B,C,D,X,MINP,pE);
        double s1 = S(A,B,C,D,X,MAXP,pE);
        pA = ( s0 < s1 ? MAXP : MINP );
      }
    }
    else  {
      pA = -b/(2*a);
    }
  }
  return pA;
}

double ScoreHQ::rootpE(double A, double B, double C, double D, double X, double pA)  {

  double a = -(C+D+X)*(1-2*pA);
  double b = C*(1-3*pA) - D*pA + X*(1-2*pA);
  double c = C*pA;
  double pE;

  if ( fabs(a) < VERY_SMALL && fabs(b) < VERY_SMALL )  {
    errflag = 1;
    return(-1);
  }

  if ( fabs(a) < VERY_SMALL)  {
    pE = -(c/b);
  }
  else  { 
    double disc = b*b-4*a*c;
    if ( disc < 0 )  {
      /* check endpoints for maximum */
      double s0 = S(A,B,C,D,X,pA,MINP);
      double s1 = S(A,B,C,D,X,pA,MAXP);
      pE = ( s0 < s1 ? MAXP : MINP );
    }
    else if ( disc > 0 )  {

      double r1 = (-b+sqrt(disc))/(2*a);
      double r2 = (-b-sqrt(disc))/(2*a);

      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP  > r2 || r2  > MAXP ) )  {
        pE = r1;
      }
      if ( ( MINP <= r2 && r2 <= MAXP ) && 
           ( MINP  > r1 || r1  > MAXP ) )  {
        pE = r2;
      }
      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP <= r2 && r2 <= MAXP ) )  {
        double d1 = d2S(C,D,X,r1,pE);
        double d2 = d2S(C,D,X,r2,pE);
        if ( d1 < 0 && d2 >= 0 )  {
          pE = r1;
        }
        else if ( d2 < 0 && d1 >= 0 )  {
          pE = r2;
        }
        else  {
          double s0 = S(A,B,C,D,X,pA,MINP);
          double s1 = S(A,B,C,D,X,pA,MAXP);
          pE = ( s0 < s1 ? MAXP : MINP );
        }
      }
      if ( ( MINP > r1 || r1 > MAXP ) && 
           ( MINP > r2 || r2 > MAXP ) )  {
        double s0 = S(A,B,C,D,X,pA,MINP);
        double s1 = S(A,B,C,D,X,pA,MAXP);
        pE = ( s0 < s1 ? MAXP : MINP );
      }
    }
    else  {
      pE = -b/(2*a);
    }
  }
  return pE;
}

double ScoreHQ::initpA(void)  {
  double numA, denA; 
  /* count A alleles */
  denA = 2*( XAARR + XAARV + XAAVV ) + 
           ( XAERR + XAEVV + XAERV );
  /* count AR-alleles */
  numA = 2*( XAARR ) + 
           ( XAARV + XAERR ) +
       0.5*( XAERV );
  return ( numA/denA );
}

double ScoreHQ::initpE(void)  {
  double numE, denE;
  /* count E alleles */
  denE = 2*( XEERR + XEERV + XEEVV ) + 
           ( XAERR + XAEVV + XAERV );

  /* count ER-alleles */
  numE = 2*( XEERR ) + 
           ( XEERV + XAERR ) +
       0.5*( XAERV );
  return ( numE/denE );
}

inline double ScoreHQ::S(double A, double B, double C, double D, double X, 
       double pA, double pE)  {
  return( A*log(pA) + B*log(1-pA) + C*log(pE) + D*log(1-pE) +
    X*log( pA + pE - 2*pA*pE ) );
}

inline double ScoreHQ::d2S(double A, double B, double X, double pA, double pE)  {
  return( -A/(pA*pA) + B/((1-pA)*(1-pA)) - 
    X*(1-2*pA)/( (pA+pE-2*pA*pE)*(pA+pE-2*pA*pE) ) );
}






/* ------------------------------------------------------------------- */
string ScoreHQ::errorString(void)  {
  ostringstream ss;
  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << XAARR;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << XAERR;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << XEERR;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << XAARV;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << XAERV;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << XEERV;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << XAAVV;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << XAEVV;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << XEEVV;
  ss << endl;
  return ss.str();
}


/* ------------------------------------------------------------------- */
void ScoreHQ::setEffectSizes(double e1, double e2)  {
  ostringstream r1str;
  r1str << fixed << setw(10) << setprecision(4) << e1;
  effectsize1 = r1str.str();
    
  ostringstream r2str;
  r2str << fixed << setw(10) << setprecision(4) << e2;
  effectsize2 = r2str.str();
}














