#include "ScoreQSUM.h"

ScoreQSUM::ScoreQSUM(char *paramfile)  {
  ostringstream errstream;
	string str;
  errstream << "[ScoreQSUM::ScoreQSUM]  ";
 
  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec = new double[nsamples];
  globalAnc = new double[nsamples];
  genoDosages = new double[nsamples];

  // geno_is_dosage=ioh.getBoolParam("dosages");

  ioh.getParam("genoformat",&genoform);
  use_adjustment = ioh.getBoolParam("adjustment"); 

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

}

ScoreQSUM::~ScoreQSUM(void)  {

  if ( outputStream.is_open() )       {  outputStream.close();  }
  if ( genoStream.is_open() )         {  genoStream.close();    }
  if ( phenoStream.is_open() )        {  phenoStream.close();   }
  if ( localAncStream.is_open() )     {  localAncStream.close();   }

  delete dipGenoVec;
  delete genoDosages;
  delete dipPhenoVec;
  delete dipLocalAnc;
}


int ScoreQSUM::evaluate(void)  {
  string str;

  ostringstream errstream;
  errstream << "[ScoreQSUM::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file");

  try {  ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec);  }
  catch ( Exception& e )  {
    errstream << endl << e.getMessage() << endl;
    throw( Exception(errstream.str()) );
  }

  ioh.getParam("thetafile", &str);
  try  {  ioh.readOneColumnFile(str, nsamples, &globalAnc);  }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage() << endl;
    throw( Exception(errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {
    if( !genoform.compare("dosage") ){
      geno_is_dosage = true;
      try { ioh.readRealLine(&genoStream, nsamples, &genoDosages); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }
    else if ( !genoform.compare("prob") )  {
      geno_is_dosage = true;
      double *tempVec = new double[2*nsamples];
      try { ioh.readRealLine(&genoStream, 2*nsamples, &tempVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
      for(int i=0;i<nsamples;i++)  {
        double p2 = tempVec[2*i];
        double p1 = tempVec[2*i+1];
        genoDosages[i] = 2.0*p2 + p1;
      }
      delete tempVec;
    }
    else  {    // genoformat is "integer"
      geno_is_dosage = false;
      try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage() << endl;
      throw( Exception(errstream.str()) );
    }

    double s2q, a2q;

    try  {
      if(geno_is_dosage){
	 ScoreS2Q S2Q(nsamples, genoDosages, dipPhenoVec, dipLocalAnc);
         s2q = S2Q.evalS2Q();
      }
      else  {
        ScoreS2Q S2Q(nsamples, dipGenoVec, dipPhenoVec, dipLocalAnc);
        s2q = S2Q.evalS2Q();
      }
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score S1S" << endl;
      errstream << e.getMessage();

      cerr << errstream.str();
      if ( output_to_file )  {
        outputStream << "no-score" << endl;
      }
      continue;
    }

    try  {
      ScoreA2Q A2Q(nsamples, dipPhenoVec, dipLocalAnc, globalAnc);
      a2q = A2Q.evalA2Q();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score A1" << endl;
      errstream << e.getMessage();

      cerr << errstream.str();
      if ( output_to_file )  {
        outputStream << "no-score" << endl;
      }
      continue;
    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << (s2q+a2q) << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << (s2q+a2q) << endl;
    }
  }
  return 0;
}

