#include "ScoreS1S.h"

ScoreS1S::ScoreS1S(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreS1S::ScoreS1S]  ";

  // read parameter file
  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  genoDosages = new double[nsamples];
  geno_is_dosage = ioh.getBoolParam("dosages");

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new int[nsamples];

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  if ( ioh.getBoolParam("oddsratio") )  {
    ioh.openOutput("orfile",&oddsratioStream);
  }

  return;
}

ScoreS1S::~ScoreS1S(void)  {

  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( oddsratioStream.is_open() )  {  oddsratioStream.close();  }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )   {  localAncStream.close();  }

  // close file streams and free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete dipLocalAnc;
}

int ScoreS1S::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreS1S::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 
  bool output_oddsratios = ioh.getBoolParam("oddsratio");

  try { ioh.readIntLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    if(geno_is_dosage){
      try { ioh.readRealLine(&genoStream, nsamples, &genoDosages); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
      
    }else{
      try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalS1S();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_oddsratios )  {
      oddsratioStream << fixed << setw(10) << 
        oddsratio1 << "  " << oddsratio2 << endl;
    }

  }

  return 0;
}

#define TINY 0.1
#define MINP 0.0001
#define MAXP 0.9999
#define RMIN 0.1
#define RMAX 10.0
#define VERY_SMALL 1e-8
#define MAX_ITER 40


double ScoreS1S::evalS1S(void)  {

  ostringstream errstream;
  errstream << "[ScoreS1S::evalS1S]  ";

  int done, itercount;
  double A0,B0,C0,D0,X0;
  double AT,BT,CT,DT,XT;
  double AS,BS,CS,DS,XS;

  string str;
  countAll(dipGenoVec, dipPhenoVec, dipLocalAnc);

  double pA = initpA();
  double pE = initpE();

  A0 = 2*(XAARRT+XAARRS) + (XAARVT+XAARVS) + (XAERRT+XAERRS);
  B0 = 2*(XAAVVT+XAAVVS) + (XAARVT+XAARVS) + (XAEVVT+XAEVVS);
  C0 = 2*(XEERRT+XEERRS) + (XAERRT+XAERRS) + (XEERVT+XEERVS);
  D0 = 2*(XEEVVT+XEEVVS) + (XEERVT+XEERVS) + (XAEVVT+XAEVVS);
  X0 = XAERVT+XAERVS;

  done = itercount = 0;
  while ( !done && itercount++ < MAX_ITER )  {
    double oldpA = pA, oldpE = pE;
    pA = rootpA(A0,B0,C0,D0,X0,pE);
    pE = rootpE(A0,B0,C0,D0,X0,pA);
    done = ( fabs(pA-oldpA) < VERY_SMALL &&
             fabs(pE-oldpE) < VERY_SMALL );
  }
  if ( itercount > MAX_ITER )  {
    errstream << "Failure to converge (null score)";
    throw( Exception(errstream.str()) );
  }
  double nullscore = S(A0,B0,C0,D0,X0,pA,pE);

  double pAT = initpAT();
  double pAS = initpAS();
  double pET = initpET();
  double pES = initpES();

  AT = 2*XAARRT + XAARVT + XAERRT;
  BT = 2*XAAVVT + XAARVT + XAEVVT;
  CT = 2*XEERRT + XAERRT + XEERVT;
  DT = 2*XEEVVT + XEERVT + XAEVVT;
  XT = XAERVT;

  AS = 2*XAARRS + XAARVS + XAERRS;
  BS = 2*XAAVVS + XAARVS + XAEVVS;
  CS = 2*XEERRS + XAERRS + XEERVS;
  DS = 2*XEEVVS + XEERVS + XAEVVS;
  XS = XAERVS;

  done = itercount = 0;
  while ( !done && itercount++ < MAX_ITER )  {
    double oldpAT = pAT, oldpET = pET;
    pAT = rootpA(AT,BT,CT,DT,XT,pET);
    pET = rootpE(AT,BT,CT,DT,XT,pAT);
    done = ( fabs(pAT-oldpAT) < VERY_SMALL &&
             fabs(pET-oldpET) < VERY_SMALL );
  }
  if ( itercount > MAX_ITER )  {
    errstream << "Failure to converge (causal score, controls)";
    throw( Exception(errstream.str()) );
  }

  done = itercount = 0;
  while ( !done && itercount++ < MAX_ITER )  {
    double oldpAS = pAS, oldpES = pES;
    pAS = rootpA(AS,BS,CS,DS,XS,pES);
    pES = rootpE(AS,BS,CS,DS,XS,pAS);
    done = ( fabs(pAS-oldpAS) < VERY_SMALL &&
             fabs(pES-oldpES) < VERY_SMALL );
  }
  if ( itercount > MAX_ITER )  {
    errstream << "Failure to converge (causal score, cases)";
    throw( Exception(errstream.str()) );
  }

  double causalscore = S(AT,BT,CT,DT,XT,pAT,pET) + 
                       S(AS,BS,CS,DS,XS,pAS,pES); 


  setOddsRatios(pAT, pAS, pET, pES);

  return( 2*(causalscore-nullscore) );

}

// ------------------------------------------------------------------------
void ScoreS1S::countAll(int *G, int *P, int *A)  {

  XAARRT = XAARVT = XAAVVT = XAARRS = XAARVS = XAAVVS = 0.0;
  XAERRT = XAERVT = XAEVVT = XAERRS = XAERVS = XAEVVS = 0.0;
  XEERRT = XEERVT = XEEVVT = XEERRS = XEERVS = XEEVVS = 0.0;

  for(int i=0;i<nsamples;i++)  {

    if ( P[i] == 9 || G[i] == 9 || A[i] == 9 )  
      continue;

    if ( P[i] == 0 )  {
      if ( A[i] == 0 )  {
        if ( G[i] == 0 )  
          XAAVVT += 1;
        if ( G[i] == 1 ) 
          XAARVT += 1;
        if ( G[i] == 2 ) 
          XAARRT += 1;
      }
      if ( A[i] == 1 )  {
        if ( G[i] == 0 )  
          XAEVVT += 1;
        if ( G[i] == 1 ) 
          XAERVT += 1;
        if ( G[i] == 2 ) 
          XAERRT += 1;;
      }
      if ( A[i] == 2 )  {
        if ( G[i] == 0 )  
          XEEVVT += 1;
        if ( G[i] == 1 ) 
          XEERVT += 1;
        if ( G[i] == 2 ) 
          XEERRT += 1;
      }
    }
    if ( P[i] == 1 )  {
      if ( A[i] == 0 )  {
        if ( G[i] == 0 )  
          XAAVVS += 1;
        if ( G[i] == 1 ) 
          XAARVS += 1;
        if ( G[i] == 2 ) 
          XAARRS += 1;
      }
      if ( A[i] == 1 )  {
        if ( G[i] == 0 )  
          XAEVVS += 1;
        if ( G[i] == 1 ) 
          XAERVS += 1;
        if ( G[i] == 2 ) 
          XAERRS += 1;
      }
      if ( A[i] == 2 )  {
        if ( G[i] == 0 )  
          XEEVVS += 1;
        if ( G[i] == 1 ) 
          XEERVS += 1;
        if ( G[i] == 2 ) 
          XEERRS += 1;
      }
    }
  } 

}


// ------------------------------------------------------------------------
inline double ScoreS1S::initpA(void)  {
  double numA, denA; 
  /* count A alleles */
  denA = 2*( XAARRT + XAARVT + XAAVVT + XAARRS + XAARVS + XAAVVS ) + 
           ( XAERRT + XAEVVT + XAERRS + XAEVVS + XAERVT + XAERVS );
  /* count AR-alleles */
  numA = 2*( XAARRT + XAARRS ) + 
           ( XAARVT + XAARVS + XAERRT + XAERRS ) +
       0.5*( XAERVT + XAERVS );
  return ( numA/denA );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::initpE(void)  {
  double numE, denE;
  /* count E alleles */
  denE = 2*( XEERRT + XEERVT + XEEVVT + XEERRS + XEERVS + XEEVVS ) + 
           ( XAERRT + XAEVVT + XAERRS + XAEVVS + XAERVT + XAERVS );

  /* count ER-alleles */
  numE = 2*( XEERRT + XEERRS ) + 
           ( XEERVT + XEERVS + XAERRT + XAERRS ) +
       0.5*( XAERVT + XAERVS );
  return ( numE/denE );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::initpAT(void)  {
  double numA, denA; 
  /* count A alleles in controls */
  denA = 2*( XAARRT + XAARVT + XAAVVT ) + 
           ( XAERRT + XAEVVT + XAERVT );
  /* count AR-alleles in controls */
  numA = 2*XAARRT + ( XAARVT + XAERRT ) + 0.5*( XAERVT );
  return ( numA/denA );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::initpAS(void)  {
  double numA, denA; 
  /* count A alleles in cases */
  denA = 2*( XAARRS + XAARVS + XAAVVS ) + 
           ( XAERRS + XAEVVS + XAERVS );
  /* count AR-alleles in cases */
  numA = 2*XAARRS + ( XAARVS + XAERRS ) + 0.5*( XAERVS );
  return ( numA/denA );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::initpET(void)  {
  double numE, denE;
  /* count E alleles in controls */
  denE = 2*( XEERRT + XEERVT + XEEVVT ) + 
           ( XAERRT + XAEVVT + XAERVT );

  /* count ER-alleles in controls */
  numE = 2*XEERRT + ( XEERVT + XAERRT ) + 0.5*( XAERVT );
  return ( numE/denE );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::initpES(void)  {
  double numE, denE;
  /* count E alleles in cases */
  denE = 2*( XEERRS + XEERVS + XEEVVS ) + 
           ( XAERRS + XAEVVS + XAERVS );

  /* count ER-alleles in cases */
  numE = 2*XEERRS + ( XEERVS + XAERRS ) + 0.5*( XAERVS );
  return ( numE/denE );
}

// ------------------------------------------------------------------------
double ScoreS1S::rootpA(double A, double B, double C, double D, double X, double pE)  {

  double a = -(A+B+X)*(1-2*pE);
  double b = A*(1-3*pE) - B*pE + X*(1-2*pE);
  double c = A*pE;
  double pA;

  if ( fabs(a) < VERY_SMALL && fabs(b) < VERY_SMALL )  {
    errflag = 1;
    return(-1);
  }

  if ( fabs(a) < VERY_SMALL)  {
    pA = -(c/b);
  }
  else  { 
    double disc = b*b-4*a*c;
    if ( disc < 0 )  {
      /* check endpoints for maximum */
      double s0, s1;
      s0 = S(A,B,C,D,X,MINP,pE);
      s1 = S(A,B,C,D,X,MAXP,pE);
      pA = ( s0 < s1 ? MAXP : MINP );
    }
    else if ( disc > 0 )  {

      double r1 = (-b+sqrt(disc))/(2*a);
      double r2 = (-b-sqrt(disc))/(2*a);

      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP  > r2 || r2  > MAXP ) )  {
        pA = r1;
      }
      if ( ( MINP <= r2 && r2 <= MAXP ) && 
           ( MINP  > r1 || r1  > MAXP ) )  {
        pA = r2;
      }
      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP <= r2 && r2 <= MAXP ) )  {
        double d1 = d2S(A,B,X,r1,pE);
        double d2 = d2S(A,B,X,r2,pE);
        if ( d1 < 0 && d2 >= 0 )  {
          pA = r1;
        }
        else if ( d2 < 0 && d1 >= 0 )  {
          pA = r2;
        }
        else  {
          double s0 = S(A,B,C,D,X,MINP,pE);
          double s1 = S(A,B,C,D,X,MAXP,pE);
          pA = ( s0 < s1 ? MAXP : MINP );
        }
      }
      if ( ( MINP > r1 || r1 > MAXP ) && 
           ( MINP > r2 || r2 > MAXP ) )  {
        double s0 = S(A,B,C,D,X,MINP,pE);
        double s1 = S(A,B,C,D,X,MAXP,pE);
        pA = ( s0 < s1 ? MAXP : MINP );
      }
    }
    else  {
      pA = -b/(2*a);
    }
  }
  return pA;
}

// ------------------------------------------------------------------------
double ScoreS1S::rootpE(double A, double B, double C, double D, double X, double pA)  {

  double a = -(C+D+X)*(1-2*pA);
  double b = C*(1-3*pA) - D*pA + X*(1-2*pA);
  double c = C*pA;
  double pE;

  if ( fabs(a) < VERY_SMALL && fabs(b) < VERY_SMALL )  {
    errflag = 1;
    return(-1);
  }

  if ( fabs(a) < VERY_SMALL)  {
    pE = -(c/b);
  }
  else  { 
    double disc = b*b-4*a*c;
    if ( disc < 0 )  {
      /* check endpoints for maximum */
      double s0 = S(A,B,C,D,X,pA,MINP);
      double s1 = S(A,B,C,D,X,pA,MAXP);
      pE = ( s0 < s1 ? MAXP : MINP );
    }
    else if ( disc > 0 )  {

      double r1 = (-b+sqrt(disc))/(2*a);
      double r2 = (-b-sqrt(disc))/(2*a);

      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP  > r2 || r2  > MAXP ) )  {
        pE = r1;
      }
      if ( ( MINP <= r2 && r2 <= MAXP ) && 
           ( MINP  > r1 || r1  > MAXP ) )  {
        pE = r2;
      }
      if ( ( MINP <= r1 && r1 <= MAXP ) && 
           ( MINP <= r2 && r2 <= MAXP ) )  {
        double d1 = d2S(C,D,X,r1,pE);
        double d2 = d2S(C,D,X,r2,pE);
        if ( d1 < 0 && d2 >= 0 )  {
          pE = r1;
        }
        else if ( d2 < 0 && d1 >= 0 )  {
          pE = r2;
        }
        else  {
          double s0 = S(A,B,C,D,X,pA,MINP);
          double s1 = S(A,B,C,D,X,pA,MAXP);
          pE = ( s0 < s1 ? MAXP : MINP );
        }
      }
      if ( ( MINP > r1 || r1 > MAXP ) && 
           ( MINP > r2 || r2 > MAXP ) )  {
        double s0 = S(A,B,C,D,X,pA,MINP);
        double s1 = S(A,B,C,D,X,pA,MAXP);
        pE = ( s0 < s1 ? MAXP : MINP );
      }
    }
    else  {
      pE = -b/(2*a);
    }
  }
  return pE;
}

// ------------------------------------------------------------------------
inline double ScoreS1S::S(double A, double B, double C, double D, double X, 
       double pA, double pE)  {
  return( A*log(pA) + B*log(1-pA) + C*log(pE) + D*log(1-pE) +
    X*log( pA + pE - 2*pA*pE ) );
}

// ------------------------------------------------------------------------
inline double ScoreS1S::d2S(double A, double B, double X, double pA, double pE)  {
  return( -A/(pA*pA) + B/((1-pA)*(1-pA)) - 
    X*(1-2*pA)/( (pA+pE-2*pA*pE)*(pA+pE-2*pA*pE) ) );
}


/* ------------------------------------------------------------------- */
string ScoreS1S::errorString(void)  {
  ostringstream ss;
  ss << "Controls: " << endl;
  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << XAARRT;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << XAERRT;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << XEERRT;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << XAARVT;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << XAERVT;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << XEERVT;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << XAAVVT;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << XAEVVT;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << XEEVVT;
  ss << endl;

  ss << "Cases:    " << endl;
  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << XAARRS;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << XAERRS;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << XEERRS;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << XAARVS;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << XAERVS;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << XEERVS;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << XAAVVS;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << XAEVVS;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << XEEVVS;
  ss << endl;
  return ss.str();
}

void ScoreS1S::setOddsRatios(double pAT, double pAS, double pET, double pES)  {
  if ( fabs(pAT) <= MINP || fabs(1-pAS) <= MINP )  {
    oddsratio1 = "large";
  }
  else if ( fabs(pAS) <= MINP || fabs(1-pAT) <= MINP )  {
    oddsratio1 = "large";
  }
  else  {
    double rprov = (pAS*(1-pAT))/(pAT*(1-pAS));
    if ( rprov < RMIN )  {
      oddsratio1 = "small";
    }
    else if ( rprov >= RMAX )  {
      oddsratio1 = "large";
    }
    else  {
      ostringstream rstr;
      rstr << fixed << setw(10) << setprecision(4) << 
        (pAS*(1-pAT))/(pAT*(1-pAS));
      oddsratio1 = rstr.str();
    }
  }

  if ( fabs(pET) <= MINP || fabs(1-pES) <= MINP )  {
    oddsratio2 = "large";
  }
  else if ( fabs(pES) <= MINP || fabs(1-pET) <= MINP )  {
    oddsratio2 = "large";
  }
  else  {
    double rprov = (pES*(1-pET))/(pET*(1-pES));

    if ( rprov < RMIN )  {
      oddsratio2 = "small";
    }
    else if ( rprov >= RMAX )  {
      oddsratio2 = "large";
    }
    else  {
      ostringstream rstr;
      rstr << fixed << setw(10) << setprecision(4) << rprov;
      oddsratio2 = rstr.str();
    }
  }
}





