#include "ScoreS2Q.h"

ScoreS2Q::ScoreS2Q(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreS2Q::ScoreS2Q]  ";

  // read parameter file
  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];
  genoDosages = new double[nsamples];
  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new double[nsamples];

  // geno_is_dosage = ioh.getBoolParam("dosages");
  ioh.getParam("genoformat",&genoform);
  use_adjustment = ioh.getBoolParam("adjustment");
	
  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }
  do_not_deallocate = 0;

  if ( ioh.getBoolParam("effectsize") )  {
    ioh.openOutput("esfile",&effectsizeStream);
  }

  // ioh.dbgShowParams();

  return;
}

ScoreS2Q::ScoreS2Q(int n, int *g, double *p, int *a)  {
  nsamples = n;
  dipGenoVec = g;
  dipPhenoVec = p;
  dipLocalAnc = a;
  do_not_deallocate = 1;
  geno_is_dosage = false;
}

ScoreS2Q::ScoreS2Q(int n, double *d, double *p, int *a)  {
  nsamples = n;
  genoDosages = d;
  dipPhenoVec = p;
  dipLocalAnc = a;
  geno_is_dosage = true;
  do_not_deallocate = 1;
}

ScoreS2Q::~ScoreS2Q(void)  {

  if ( do_not_deallocate ) 
    return;

  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( effectsizeStream.is_open() ) {  effectsizeStream.close(); }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )   {  localAncStream.close();   }

  // close file streams and free memory
  delete dipGenoVec;
  delete genoDosages;
  delete dipPhenoVec;
  delete dipLocalAnc;

}

int ScoreS2Q::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreS2Q::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 
  bool output_effectsize = ioh.getBoolParam("effectsize"); 

  try { ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    if( !genoform.compare("dosage") ){
      geno_is_dosage = true;
      try { ioh.readRealLine(&genoStream, nsamples, &genoDosages); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }
    else if ( !genoform.compare("prob") )  {
      geno_is_dosage = true;
      double *tempVec = new double[2*nsamples];
      try { ioh.readRealLine(&genoStream, 2*nsamples, &tempVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
      for(int i=0;i<nsamples;i++)  {
        double p2 = tempVec[2*i];
        double p1 = tempVec[2*i+1];
        genoDosages[i] = 2.0*p2 + p1;
      }
      delete tempVec;
    }
    else  {    // genoformat is "integer"
      geno_is_dosage = false;
      try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalS2Q();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_effectsize )  {
      effectsizeStream << fixed << setw(10) 
        << effectsize << endl;
    }
  }

  return 0;
}

double ScoreS2Q::evalS2Q(void)  {

  ostringstream errstream;
  errstream << "[ScoreS2Q::evalS2Q]  ";
 
  double *gamma = new double[nsamples];    // mean-adjusted genotype
  double *alpha = new double[nsamples];    // mean-adjusted localanc
  double *phi   = new double[nsamples];    // mean-adjusted phenotype
  int *baddata = new int[nsamples];

  int count;
  double sum;

  // locate bad data
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    baddata[i] = 0;
    if ( geno_is_dosage==false && dipGenoVec[i] == 9 )
      baddata[i] = 1;
	if ( geno_is_dosage==true && (genoDosages[i] <0 || genoDosages[i] >2) )
		  baddata[i] = 1;
	  
    if ( dipLocalAnc[i] == 9 )
      baddata[i] = 1;
    if ( fabs(dipPhenoVec[i]+1024.0) < 1e-6 )   // "missing"
      baddata[i] = 1;

    if ( !baddata[i] )
      count++;
  }

  // prevent division by zero
  if ( count == 0 )  {
    errstream << "no valid data";
    throw(Exception(errstream.str()));
  } 
  if(geno_is_dosage){
	
    // mean-adjust genotype
    sum = 0.0;
    count = 0;
    for(int i=0;i<nsamples;i++)  {
      if ( !baddata[i] )  {
        sum += genoDosages[i];
        count++;
      }
    }
    for(int i=0;i<nsamples;i++)  {
      if ( !baddata[i] )  {
        gamma[i] = genoDosages[i] - (sum/count);
      }
    }
  }
  else  {
    // mean-adjust genotype
    sum = 0.0;
    count = 0;
    for(int i=0;i<nsamples;i++)  {
      if ( !baddata[i] )  {
        sum += dipGenoVec[i];
	count++;
      }
    }
    for(int i=0;i<nsamples;i++)  {
      if ( !baddata[i] )  {
        gamma[i] = dipGenoVec[i] - (sum/count);
      }
    }	
  }

  // mean-adjust phenotype
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      sum += dipPhenoVec[i];
      count++;
    }
  }
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      phi[i] = dipPhenoVec[i] - (sum/count);
    }
  }

  // mean-adjust local ancestry
  sum = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      double a = dipLocalAnc[i];
      sum += a;
      count++;
    }
  }
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      double a = dipLocalAnc[i];  
      alpha[i] = a - (sum/count);
    }
  }

  // adjust genotype for ancestry
  double num = 0.0;
  double den = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      num += gamma[i]*alpha[i];
      den += alpha[i]*alpha[i];
    }
  }
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      gamma[i] -= (num/den)*alpha[i];
    }
  }

  // adjust phenotype for ancestry
  num = 0.0;
  den = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      num += phi[i]*alpha[i];
      den += alpha[i]*alpha[i];
    }
  }
  for(int i=0;i<nsamples;i++)  {
    if ( !baddata[i] )  {
      phi[i] -= (num/den)*alpha[i];
    }
  }

  // compute correlation
  int sum1 = 0;
  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    sumx  += gamma[i];
    sumy  += phi[i];
    sumxx += (gamma[i]*gamma[i]);
    sumxy += (gamma[i]*phi[i]);
    sumyy += (phi[i]*phi[i]);
    sum1  += 1;
  }
 
  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1 - mx*mx);
  double sy = sqrt(sumyy/sum1 - my*my);

  if ( sx < 1e-8 || sy < 1e-8 )  {
    errstream << "zero variance" << endl;
    throw(Exception(errstream.str()));
  }

  double corr = (sumxy/sum1 - mx*my)/(sx*sy);
  setEffectSize(baddata, gamma, phi);

  delete baddata;
  delete gamma;
  delete alpha;
  delete phi;

  return((sum1-1)*corr*corr);
}


void ScoreS2Q::setEffectSize(int *baddata, double *gamma, double *phi)  {

  double gdotg = 0.0;
  double pdotg = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    gdotg += gamma[i]*gamma[i];
    pdotg += phi[i]*gamma[i];
  }

  ostringstream rstr;
  rstr << fixed << setw(10) << setprecision(4) 
    << (pdotg/gdotg);
  effectsize = rstr.str();  
}
